#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>

#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_psi.h>

#include "../param.h"
#include "generalized_gaussian.h"
#include "common/dynamic.h"

unsigned int EF(arguments_len) = 2;
struct dynamic_type EF(arguments_descr)[2] = {
  {"mu"  , offsetof(struct family, mu  ), DOUBLE},
  {"beta", offsetof(struct family, beta), DOUBLE},
};

unsigned int EF(source_len) = 1;
struct dynamic_type EF(source_descr)[1] = {
  {"alpha", offsetof(struct source, alpha), DOUBLE},
};

unsigned int EF(natural_len) = 1;
struct dynamic_type EF(natural_descr)[1] = {
  {"theta1", offsetof(struct natural, theta1), DOUBLE},
};

unsigned int EF(expectation_len) = 1;
struct dynamic_type EF(expectation_descr)[1] = {
  {"eta1", offsetof(struct expectation, eta1), DOUBLE},
};

void EF(estimate_mu)(family* ef, unsigned int n, double* data, char* mask) {
  double mu = 0.;

  unsigned int count = 0;
  for (unsigned int j=0; j<n; j++) {
    if (mask[j]) {
      count++;

      mu += data[j];
    }
  }

  ef->mu = mu / count;
}

double EF(f_beta)(double beta, double mu, unsigned int n, double* data, char* mask) {
  double sum_x_beta_log_x = 0;
  double sum_x_beta       = 0;

  unsigned int count = 0;
  for (unsigned int i=0; i<n; i++) {
    if (mask[i]) {
      count++;

      double x      = fabs(data[i] - mu);
      double x_beta = pow(x, beta);

      sum_x_beta_log_x += x_beta * log(x);
      sum_x_beta       += x_beta;
    }
  }
  return 1 + gsl_sf_psi(1 / beta) / beta \
    - sum_x_beta_log_x / sum_x_beta + log(beta * sum_x_beta / count) / beta;
}

void EF(estimate_beta)(family* ef, unsigned int n, double* data, char* mask) {
  unsigned int count = 30;

  double mu   = ef->mu;

  double beta1 = 0.01;
  double beta2 =  20;
  double beta  = (beta1 + beta2) / 2;

  double y;
  double y1;

  for (unsigned int i=0; i<count; i++) {
    y1 = GeneralizedGaussian_f_beta(beta1, mu, n, data, mask);
    y  = GeneralizedGaussian_f_beta(beta , mu, n, data, mask);

    if (y * y1 >= 0)
      beta1 = beta;
    else
      beta2 = beta;

    beta = (beta1 + beta2) / 2;
  }

  ef->beta = beta;
}

source* EF(create_source)(family* ef, double alpha) {
  source* param = EF(new_param)(ef, SOURCE);
  param->alpha = alpha;

  return param;
}

natural* EF(create_natural)(family* ef, double theta1) {
  natural* param = EF(new_param)(ef, NATURAL);
  param->theta1 = theta1;

  return param;
}

expectation* EF(create_expectation)(family* ef, double eta1) {
  expectation* param = EF(new_param)(ef, EXPECTATION);
  param->eta1 = eta1;

  return param;
}

double EF(pdf)(family* ef, double* point, param* p) {
  CHECK_EF(ef, p);

  double x = *point;
  
  source* lambda = (source*)ef->as_source(ef, p);
  double alpha   = lambda->alpha;
  double mu      = ef->mu;
  double beta    = ef->beta;

  return beta / (2 * alpha * gsl_sf_gamma(1 / beta)) * \
    exp(- pow(fabs(x - mu) / alpha, beta));
}

void EF(lambda2theta)(family* ef, param* lambda0, param* theta0) {
  debug("lambda2theta\n");
  CHECK_EF(ef, lambda0);
  CHECK_EF(ef, theta0);
  CHECK_TYPE(lambda0, SOURCE);
  CHECK_TYPE(theta0, NATURAL);
  source* lambda = (source*)lambda0;
  natural* theta = (natural*)theta0;

  double alpha  = lambda->alpha;
  double beta   = ef->beta;

  theta->theta1 = pow(alpha, -beta);
}

void EF(theta2lambda)(family* ef, param* theta0, param* lambda0) {
  debug("theta2lambda\n");
  CHECK_EF(ef, lambda0);
  CHECK_EF(ef, theta0);
  CHECK_TYPE(lambda0, SOURCE);
  CHECK_TYPE(theta0, NATURAL);
  source* lambda = (source*)lambda0;
  natural* theta = (natural*)theta0;

  double theta1 = theta->theta1;
  double beta   = ef->beta;

  lambda->alpha  = pow(theta1, -1/beta);
}

void EF(lambda2eta)(family* ef, param* lambda0, param* eta0) {
  debug("lambda2eta\n");
  CHECK_EF(ef, lambda0);
  CHECK_EF(ef, eta0);
  CHECK_TYPE(lambda0, SOURCE);
  CHECK_TYPE(eta0, EXPECTATION);

  double beta      = ef->beta;
  double alpha     = ((source*)lambda0)->alpha;
  expectation* eta = (expectation*)eta0;

  eta->eta1 = - pow(alpha, beta) / beta;
}

void EF(eta2lambda)(family* ef, param* eta0, param* lambda0) {
  debug("eta2lambda\n");
  CHECK_EF(ef, lambda0);
  CHECK_EF(ef, eta0);
  CHECK_TYPE(lambda0, SOURCE);
  CHECK_TYPE(eta0, EXPECTATION);

  double beta    = ef->beta;
  double eta     = ((expectation*)eta0)->eta1;
  source* lambda = (source*)lambda0;

  lambda->alpha = pow(- beta * eta, 1/beta);
}

double EF(F)(family* ef, param* theta0) {
  CHECK_EF(ef, theta0);
  CHECK_TYPE(theta0, NATURAL);

  double beta  = ef->beta;
  double theta = ((natural*)theta0)->theta1;

  return -log(theta) / beta - log(beta / (2 * gsl_sf_gamma(1/beta)));
}

void EF(gradF)(family* ef, param* theta0, param* eta0) {
  CHECK_EF(ef, theta0);
  CHECK_EF(ef, eta0);
  CHECK_TYPE(theta0, NATURAL);
  CHECK_TYPE(eta0,   EXPECTATION);

  double beta      = ef->beta;
  expectation* eta = (expectation*)eta0;
  double theta     = ((natural*)theta0)->theta1;

  eta->eta1 = - 1 / (beta * theta);
}

double EF(G)(family* ef, param* eta0) {
  CHECK_EF(ef, eta0);
  CHECK_TYPE(eta0, EXPECTATION);

  double beta = ef->beta;
  double eta  = ((expectation*)eta0)->eta1;

  return - 1 / beta - log(- beta * eta) / beta + log(beta / (2 * gsl_sf_gamma(1/beta)));
}

void EF(gradG)(family* ef, param* eta0, param* theta0) {
  CHECK_EF(ef, eta0);
  CHECK_EF(ef, theta0);
  CHECK_TYPE(eta0,   EXPECTATION);
  CHECK_TYPE(theta0, NATURAL);

  double beta    = ef->beta;
  natural* theta = (natural*)theta0;
  double eta     = ((expectation*)eta0)->eta1;

  theta->theta1 = - 1 / (beta * eta);
}

void EF(t)(family* ef, double *point, param* eta0) {
  CHECK_EF(ef, eta0);
  CHECK_TYPE(eta0, EXPECTATION);
  expectation* eta = (expectation*)eta0;
  double mu        = ef->mu;
  double beta      = ef->beta;
  double x         = *point;

  eta->eta1 = - pow(fabs(x - mu), beta);
}

double EF(k)(family* ef, double x) {
  return 0;
}

double* EF(rand)(family* ef, param* param0, gsl_rng *rng) {
  return 0;
}

#define CONSTRUCTOR
#include "common/all.c"

family* NAME(double mu, double beta) {
  family* ef = talloc(NULL, struct family);

  ef->name           = __S(NAME);
  ef->dim            = 1;
  ef->pdf            = &EF(pdf           );
  ef->iter_pdf       = &EF(iter_pdf      );
  ef->lambda2theta   = &EF(lambda2theta  );
  ef->theta2lambda   = &EF(theta2lambda  );
  ef->lambda2eta     = &EF(lambda2eta    );
  ef->eta2lambda     = &EF(eta2lambda    );
  ef->theta2eta      = &EF(theta2eta     );
  ef->eta2theta      = &EF(eta2theta     );
  ef->fprint         = &EF(fprint        );
  ef->F              = &EF(F             );
  ef->gradF          = &EF(gradF         );
  ef->G              = &EF(G             );
  ef->gradG          = &EF(gradG         );
  ef->mle            = &EF(mle           );
  ef->new_param      = &EF(new_param     );
  ef->delete_param   = &EF(delete_param  );
  ef->copy_param     = &EF(copy_param    );
  ef->t              = &EF(t             );
  ef->add            = &EF(add           );
  ef->scale          = &EF(scale         );
  ef->scalar         = &EF(scalar        );
  ef->zero           = &EF(zero          );
  ef->rand           = &EF(rand          );
  ef->as_source      = &EF(as_source     );
  ef->as_natural     = &EF(as_natural    );
  ef->as_expectation = &EF(as_expectation);

  ef->mu   = mu;
  ef->beta = beta;

  return ef;
}

